import gzip
import os
import six
import warnings
import pandas as pd
from io import TextIOWrapper

try:
    import fastavro as avro

    avro_imported = True
except ImportError:
    avro_imported = False


def read_df(datafile, **read_args):
    """ Read a dataframe from file based on the file extension

    The following formats are supported:
    parquet, avro, csv, pickle

    Parameters
    ----------
    datafile : DataFile
        DataFile instance with the path and handle
    read_args : optional
        All keyword args are passed to the read function

    Returns
    -------
    data : pd.DataFrame

    Notes
    -----
    The read functions are taken from pandas, e.g. pd.read_csv
    Check the pandas doc for more information on the supported arguments

    """
    filetype, compression = _get_extension(datafile.path)
    reader = _readers[filetype]
    if (
        reader in (pd.read_csv, pd.read_pickle, pd.read_json)
        and compression is not None
    ):
        read_args["compression"] = compression
    elif reader in (pd.read_csv, pd.read_pickle, pd.read_json):
        read_args[
            "compression"
        ] = None  # default "infer" incompatible with handles as of 0.24
    if reader == pd.read_json:
        # Default json file is newline delimited json records, but can be overwritten
        defaults = {"lines": True, "orient": "records"}
        defaults.update(read_args)
        read_args = defaults

    with datafile.handle("rb") as f:
        return reader(f, **read_args)


def write_df(df, datafile, **write_args):
    """ Write a dataframe to file based on the file extension

    The following formats are supported:
    parquet, avro, csv, pickle

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to write to disk
    datafile : DataFile
        Datafile instance with the path and file handle
    write_args : optional
        All keyword args are passed to the write function

    Notes
    -----
    The write functions are taken from pandas, e.g. pd.to_csv
    Check the pandas doc for more information on the supported arguments

    """
    extension, compression = _get_extension(datafile.path)
    write_name = _writers[extension]
    # infer compression from filepath or from explicit arg
    compression = compression or write_args.get("compression")
    with datafile.handle("wb") as buf:
        # Some customizations for different file types
        if write_name == "to_avro":
            return _write_avro(df, buf, **write_args)

        if (
            write_name == "to_parquet"
            and not pd.Series(df.columns).map(type).eq(str).all()
        ):
            warnings.warn(
                "Dataframe contains non-string column names, which cannot be saved in parquet.\n"
                "Blocks will attempt to convert them to strings."
            )
            df.columns = df.columns.astype("str")

        if write_name == "to_json":
            defaults = {"lines": True, "orient": "records"}
            defaults.update(write_args)
            write_args = defaults

        if write_name == "to_csv":
            # make index=False the default for similar behaviour to other formats
            write_args["index"] = write_args.get("index", False)

        # For csv and pickle we have to manually compress
        manual_compress = write_name in ("to_csv", "to_pickle", "to_json")
        if manual_compress:
            write_args[
                "compression"
            ] = None  # default "infer" incompatible with handles as of 0.24

        if manual_compress and compression == "gzip":
            buf = gzip.GzipFile(fileobj=buf, mode="w")
        elif manual_compress and compression is not None:
            raise ValueError(
                "Compression {} is not supported for CSV/Pickle".format(compression)
            )

        # And for 23 compatibility need a textiowrapper for text formats
        if write_name in ("to_csv", "to_json") and six.PY3:
            buf = TextIOWrapper(buf)

        write_fn = getattr(df, write_name)
        write_fn(buf, **write_args)
        buf.close()


def _read_avro(handle, **read_args):
    if not avro_imported:
        raise ImportError(
            "Avro support requires fastavro.\n"
            "Install blocks with the [avro] option or `pip install fastavro`"
        )
    records = []
    avro_reader = avro.reader(handle)
    for record in avro_reader:
        records.append(record)
    return pd.DataFrame.from_dict(records)


def _write_avro(df, handle, **write_args):
    if not avro_imported:
        raise ImportError(
            "Avro support requires fastavro.\n"
            "Install blocks with the [avro] option or `pip install fastavro`"
        )
    schema = None
    schema_path = None
    try:
        schema = write_args["schema"]
    except KeyError:
        try:
            schema_path = write_args["schema_path"]
        except KeyError:
            raise Exception(
                "You must provide a schema or schema path when writing to Avro"
            )
    if schema is None:
        schema = avro.schema.load_schema(schema_path)
    records = df.to_dict("records")
    avro.writer(handle, schema, records)


def _get_extension(path):
    name, ext = os.path.splitext(path)
    # Support compression extensions, eg part.csv.gz
    comp = None
    if ext in _compressions and "." in name:
        comp = ext
        name, ext = os.path.splitext(name)
    return ext, _compressions.get(comp)


_readers = {
    ".pq": pd.read_parquet,
    ".parquet": pd.read_parquet,
    ".csv": pd.read_csv,
    ".pkl": pd.read_pickle,
    ".avro": _read_avro,
    ".json": pd.read_json,
}


_writers = {
    ".pq": "to_parquet",
    ".parquet": "to_parquet",
    ".csv": "to_csv",
    ".pkl": "to_pickle",
    ".avro": "to_avro",
    ".json": "to_json",
}

_compressions = {".gz": "gzip", ".bz2": "bz2", ".zip": "zip", ".xz": "xz"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from blocks import core


def test_iterate(populated):
    dfs = core.iterate(populated)
    for i in range(4):
        for j in range(4):
            cname, rname, df = next(dfs)
            assert cname == "c{}".format(i)
            assert rname == "part.{}.csv".format(j)
            assert df.shape == (10, 11)


def test_iterate_ordered(populated):
    order = ["c3", "c1", "c2", "c0"]
    dfs = core.iterate(populated, cgroups=order)
    for i in range(4):
        for j in range(4):
            cname, rname, df = next(dfs)
            assert cname == order[i]
            assert rname == "part.{}.csv".format(j)
            assert df.shape == (10, 11)


def test_iterate_axis0(populated):
    dfs = core.iterate(populated, axis=0)
    for i in range(4):
        rname, df = next(dfs)
        assert rname == "part.{}.csv".format(i)
        assert df.shape == (10, 41)


def test_iterate_axis1(populated):
    dfs = core.iterate(populated, axis=1)
    for i in range(4):
        cname, df = next(dfs)
        assert cname == "c{}".format(i)
        assert df.shape == (40, 11)



import os
import pytest
import pandas as pd
import numpy as np

from blocks.filesystem import GCSFileSystem, GCSNativeFileSystem
from delegator import run

BUCKET = "gs://blocks-example"

if os.environ.get("CI"):
    inputs = ["local"]
    outputs = ["local"]
    temps = ["local"]
    filesystems = ["gcs"]
else:
    inputs = ["local", "gcs", "gcs_extra"]
    outputs = ["local", "gcs"]
    temps = ["local", "gcs"]
    filesystems = ["gcs", "native"]


@pytest.fixture(scope="session", params=filesystems)
def fs(request):
    if request.param == "gcs":
        return GCSFileSystem()

    if request.param == "native":
        return GCSNativeFileSystem()


@pytest.fixture(scope="function", params=temps)
def temp(request, tmpdir_factory):
    if request.param == "local":
        path = str(tmpdir_factory.mktemp("temp"))
        yield path

    if request.param == "gcs":
        path = os.path.join(BUCKET, "temp")
        yield path
        run("gsutil -m rm -r {}".format(path))


@pytest.fixture(scope="session")
def populated_local(request, tmpdir_factory):
    tmpdir = str(tmpdir_factory.mktemp("data"))
    _populate(tmpdir)
    return tmpdir


# This is the same directory structure as above but paramatrized on different file systems
@pytest.fixture(scope="session", params=inputs)
def populated(request, populated_local):
    if request.param == "local":
        yield populated_local

    if request.param == "gcs":
        path = os.path.join(BUCKET, "data1")
        run("gsutil -m cp -r {} {}".format(populated_local, path))
        yield path
        run("gsutil -m rm -r {}".format(path))

    if request.param == "gcs_extra":
        path = os.path.join(BUCKET, "data2")
        run("gsutil -m cp -r {} {}".format(populated_local, path))
        # Also add an extra file
        run("touch extra")
        run("gsutil cp extra {}".format(path))
        os.remove("extra")
        yield path
        run("gsutil -m rm -r {}".format(path))


@pytest.fixture(scope="session")
def keys():
    return pd.Series(["key{:02d}".format(i) for i in range(40)])


@pytest.fixture()
def randomdata():
    df = pd.DataFrame(
        np.random.rand(10, 10), columns=["f{}".format(i) for i in range(10)]
    )
    return df


@pytest.fixture()
def datadir_local(request, tmpdir_factory):
    return str(tmpdir_factory.mktemp("data"))


@pytest.fixture(params=outputs)
def datadir(request, tmpdir_factory):
    if request.param == "local":
        tmpdir = str(tmpdir_factory.mktemp("data"))
        yield tmpdir

    if request.param == "gcs":
        path = os.path.join(BUCKET, "output")
        yield path
        run("gsutil -m rm -r {}".format(path))


def _populate(tmpdir):
    """ Create a directory of blocks with 4 cgroups and 4 rgroups
    """
    for c in range(4):
        cgroup = os.path.join(tmpdir, "c{}".format(c))
        if not os.path.exists(cgroup):
            os.makedirs(cgroup)
        for r in range(4):
            df = pd.DataFrame(
                np.random.rand(10, 10),
                index=list(range(r * 10, (r + 1) * 10)),
                columns=["f{}_{}".format(c, i) for i in range(10)],
            )
            df["key"] = [
                "key{:02d}".format(i) for i in df.index
            ]  # common key for merges
            df.to_csv(os.path.join(cgroup, "part.{}.csv".format(r)), index=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from uuid import uuid4

TEST_STRING = b"test"


def test_ls_directory(populated, fs):
    ex = os.path.join(populated, "")
    expected = [os.path.join(populated, "c{}".format(i)) for i in range(4)]

    # one of the fixtures has an extra file
    found = fs.ls(ex)
    if len(found) > 4:
        found = found[:-1]
    assert found == expected


def test_ls_wildcard(populated, fs):
    ex = os.path.join(populated, "*/part.1.csv")
    expected = [os.path.join(populated, "c{}/part.1.csv".format(i)) for i in range(4)]
    assert fs.ls(ex) == expected


def test_ls_double_wildcard(populated, fs):
    ex = os.path.join(populated, "**")
    expected = {
        os.path.join(populated, "c{}/part.{}.csv".format(i, j))
        for i in range(4)
        for j in range(4)
    }
    assert expected.issubset(set(fs.ls(ex)))


def test_ls_pattern(populated, fs):
    ex = os.path.join(populated, "*/part.[01].csv")
    expected = [
        os.path.join(populated, "c{}/part.{}.csv".format(i, j))
        for i in range(4)
        for j in range(2)
    ]
    assert fs.ls(ex) == expected


def test_open_read(populated, fs):
    with fs.open(os.path.join(populated, "c0/part.0.csv"), "r") as f:
        assert f.readline() == "f0_0,f0_1,f0_2,f0_3,f0_4,f0_5,f0_6,f0_7,f0_8,f0_9,key\n"


def test_open_write(temp, fs):
    content = str(uuid4())
    path = os.path.join(temp, "content")
    with fs.open(path, "w") as f:
        f.write(content)

    with fs.open(path, "r") as f:
        assert f.read() == content


def test_store_access(temp, fs):
    paths = []
    with fs.store(temp, ["ex1.txt", "ex2.txt"]) as datafiles:
        for d in datafiles:
            paths.append(d.path)
            with d.handle("wb") as f:
                f.write(TEST_STRING)

    datafiles = fs.access(paths)
    for d in datafiles:
        with d.handle("rb") as f:
            assert f.read() == TEST_STRING


def test_copy_recursive_to_local(populated, tmpdir, fs):
    dest = str(tmpdir)
    fs.cp(populated, dest, recursive=True)
    source = [p.replace(populated, "") for p in fs.ls(populated + "/**")]
    copy = fs.ls(dest + "/**")
    assert (s in c for s, c in zip(source, copy))


def test_copy_recursive_matched(populated, fs):
    dest = populated.replace("data", "copy")
    fs.cp(populated, dest, recursive=True)
    source = [p.replace(populated, "") for p in fs.ls(populated + "/**")]
    copy = fs.ls(dest + "/**")
    assert (s in c for s, c in zip(source, copy))


def test_rm(populated, fs):
    dest = populated.replace("data", "copy")
    fs.cp([os.path.join(populated, "")], dest, recursive=True)
    assert fs.ls(dest)
    fs.rm([dest], recursive=True)
    assert fs.ls(dest) == []



import os
import blocks


def test_pickle(fs, temp):
    original = {"example": 0.0}
    path = os.path.join(temp, "test.pkl")
    blocks.pickle(original, path, filesystem=fs)
    unpickled = blocks.unpickle(path, filesystem=fs)
    assert original == unpickled



import numpy as np
import blocks


def test_divide(datadir, randomdata, fs):
    blocks.divide(randomdata, datadir, 10, extension=".pkl")
    assert len(fs.ls(datadir)) == 10
    df = blocks.assemble(datadir)
    assert np.isclose(df, randomdata).all().all()


def test_divide_offset(datadir, randomdata, fs):
    blocks.divide(randomdata, datadir, 10, extension=".pkl")
    blocks.divide(randomdata, datadir, 10, 10, extension=".pkl")
    assert len(fs.ls(datadir)) == 20
    df = blocks.assemble(datadir)
    expected = randomdata.append(randomdata)
    assert np.isclose(df, expected).all().all()


def test_divide_cgroups(datadir, randomdata, fs):
    randomdata.insert(0, "key", list(range(10)))
    cgroups_columns = {
        "cgroup1": ["key", "f0", "f1", "f2"],
        "cgroup2": ["key", "f3", "f4", "f5"],
        "cgroup3": ["key", "f6", "f7", "f8", "f9"],
    }

    blocks.divide(
        randomdata, datadir, 10, cgroup_columns=cgroups_columns, extension=".pkl"
    )
    assert len(fs.ls(datadir)) == 3
    for subdir in fs.ls(datadir):
        assert len(fs.ls(subdir)) == 10
    df = blocks.assemble(datadir)
    assert df.equals(randomdata)



# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
import warnings
import pickle as _pickle

from functools import reduce
from collections import defaultdict, OrderedDict
from blocks.dfio import read_df, write_df
from blocks.filesystem import GCSFileSystem, GCSNativeFileSystem


def assemble(
    path,
    cgroups=None,
    rgroups=None,
    read_args={},
    cgroup_args={},
    merge="inner",
    filesystem=GCSFileSystem(),
):
    """ Assemble multiple dataframe blocks into a single frame

    Each file included in the path (or subdirs of that path) is combined into
    a single dataframe by first concatenating over row groups and then merging
    over cgroups. The merges are performed in the order of listed cgroups if
    provided, otherwise in alphabetic order. Files are opened by a method inferred
    from their extension

    Parameters
    ----------
    path : str
        The glob-able path to all datafiles to assemble into a frame
        e.g. gs://example/*/*, gs://example/*/part.0.pq, gs://example/c[1-2]/*
        See the README for a more detailed explanation
    cgroups : list of str, optional
        The list of cgroups (folder names) to include from the glob path
    rgroups : list of str, optional
        The list of rgroups (file names) to include from the glob path
    read_args : optional
        Any additional keyword args to pass to the read function
    cgroup_args : {cgroup: kwargs}, optional
        Any cgroup specific read arguments, where each key is the name
        of the cgroup and each value is a dictionary of keyword args
    merge : one of 'left', 'right', 'outer', 'inner', default 'inner'
        The merge strategy to pass to pandas.merge
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API

    Returns
    -------
    data : pd.DataFrame
        The combined dataframe from all the blocks

    """
    grouped = _collect(path, cgroups, rgroups, filesystem)

    # ----------------------------------------
    # Concatenate all rgroups
    # ----------------------------------------
    frames = []

    for group in grouped:
        datafiles = grouped[group]
        args = read_args.copy()
        if group in cgroup_args:
            args.update(cgroup_args[group])
        frames.append(pd.concat(read_df(d, **args) for d in datafiles))

    # ----------------------------------------
    # Merge all cgroups
    # ----------------------------------------
    df = _merge_all(frames, merge=merge)

    # ----------------------------------------
    # Delete temporary files
    # ----------------------------------------
    for file in datafiles:
        if hasattr(file.handle, "name"):
            tmp_file_path = file.handle.name
            if os.path.exists(tmp_file_path):
                os.remove(file.handle.name)
    return df


def iterate(
    path,
    axis=-1,
    cgroups=None,
    rgroups=None,
    read_args={},
    cgroup_args={},
    merge="inner",
    filesystem=GCSFileSystem(),
):
    """ Iterate over dataframe blocks

    Each file include in the path (or subdirs of that path) is opened as a
    dataframe and returned in a generator of (cname, rname, dataframe).
    Files are opened by a method inferred from their extension

    Parameters
    ----------
    path : str
        The glob-able path to all datafiles to assemble into a frame
        e.g. gs://example/*/*, gs://example/*/part.0.pq, gs://example/c[1-2]/*
        See the README for a more detailed explanation
    axis : int, default -1
        The axis to iterate along
        If -1 (the default), iterate over both columns and rows
        If 0, iterate over the rgroups, combining any cgroups
        If 1, iterate over the cgroups, combining any rgroups
    cgroups : list of str, or {str: args} optional
        The list of cgroups (folder names) to include from the glob path
    rgroups : list of str, optional
        The list of rgroups (file names) to include from the glob path
    read_args : dict, optional
        Any additional keyword args to pass to the read function
    cgroup_args : {cgroup: kwargs}, optional
        Any cgroup specific read arguments, where each key is the name
        of the cgroup and each value is a dictionary of keyword args
    merge : one of 'left', 'right', 'outer', 'inner', default 'inner'
        The merge strategy to pass to pandas.merge, only used when axis=0
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API

    Returns
    -------
    data : generator
        A generator of (cname, rname, dataframe) for each collected path
        If axis=0, yields (rname, dataframe)
        If axis=1, yields (cname, dataframe)

    """
    grouped = _collect(path, cgroups, rgroups, filesystem)

    if axis == -1:
        for cgroup in grouped:
            args = read_args.copy()
            if cgroup in cgroup_args:
                args.update(cgroup_args[cgroup])
            for datafile in grouped[cgroup]:
                yield _cname(datafile.path), _rname(datafile.path), read_df(
                    datafile, **args
                )

    elif axis == 0:
        # find the shared files among all subfolders
        rgroups = _shared_rgroups(grouped)

        for rgroup in sorted(rgroups):
            frames = []
            for cgroup in grouped:
                datafile = next(
                    d for d in grouped[cgroup] if os.path.basename(d.path) == rgroup
                )

                args = read_args.copy()
                if cgroup in cgroup_args:
                    args.update(cgroup_args[cgroup])
                frames.append(read_df(datafile, **args))
            yield rgroup, _merge_all(frames, merge=merge)

    elif axis == 1:
        for cgroup in grouped:
            datafiles = grouped[cgroup]
            args = read_args.copy()
            if cgroup in cgroup_args:
                args.update(cgroup_args[cgroup])
            yield cgroup, pd.concat(read_df(d, **args) for d in datafiles)

    else:
        raise ValueError("Invalid choice for axis, options are -1, 0, 1")


def partitioned(
    path,
    cgroups=None,
    rgroups=None,
    read_args={},
    cgroup_args={},
    merge="inner",
    filesystem=GCSFileSystem(),
):
    """ Return a partitioned dask dataframe, where each partition is a row group

    The results are the same as iterate with axis=0, except that it returns a dask dataframe
    instead of a generator. Note that this requires dask to be installed

    Parameters
    ----------
    path : str
        The glob-able path to all datafiles to assemble into a frame
        e.g. gs://example/*/*, gs://example/*/part.0.pq, gs://example/c[1-2]/*
        See the README for a more detailed explanation
    cgroups : list of str, or {str: args} optional
        The list of cgroups (folder names) to include from the glob path
    rgroups : list of str, optional
        The list of rgroups (file names) to include from the glob path
    read_args : dict, optional
        Any additional keyword args to pass to the read function
    cgroup_args : {cgroup: kwargs}, optional
        Any cgroup specific read arguments, where each key is the name
        of the cgroup and each value is a dictionary of keyword args
    merge : one of 'left', 'right', 'outer', 'inner', default 'inner'
        The merge strategy to pass to pandas.merge, only used when axis=0
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API

    Returns
    -------
    data : dask.dataframe
        A dask dataframe partitioned by row groups, with all cgroups merged

    """
    try:
        import dask
        import dask.dataframe as dd
    except ImportError:
        raise ImportError("Partitioned requires dask[dataframe] to be installed")

    grouped = _collect(path, cgroups, rgroups, filesystem)
    blocks = []

    @dask.delayed()
    def merged(rgroup):
        frames = []
        for cgroup in grouped:
            datafile = next(
                d for d in grouped[cgroup] if os.path.basename(d.path) == rgroup
            )
            args = read_args.copy()
            if cgroup in cgroup_args:
                args.update(cgroup_args[cgroup])
            frames.append(read_df(datafile, **args))
        return _merge_all(frames, merge=merge)

    for rgroup in _shared_rgroups(grouped):
        blocks.append(merged(rgroup))
    return dd.from_delayed(blocks)


def place(df, path, filesystem=GCSFileSystem(), **write_args):
    """ Place a dataframe block onto the filesystem at the specified path

    Parameters
    ----------
    df : pd.DataFrame
        The data to place
    path : str
        Path to the directory (possibly on GCS) in which to place the columns
    write_args : dict
        Any additional args to pass to the write function
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API

    """
    bucket, fname = os.path.dirname(path), os.path.basename(path)
    with filesystem.store(bucket, [fname]) as datafiles:
        write_df(df, datafiles[0], **write_args)


def divide(
    df,
    path,
    n_rgroup=1,
    rgroup_offset=0,
    cgroup_columns=None,
    extension=".pq",
    convert=False,
    filesystem=GCSFileSystem(),
    prefix=None,
    **write_args
):
    """ Split a dataframe into rgroups/cgroups and save to disk

    Note that this splitting does not preserve the original index, so make sure
    to have another column to track values

    Parameters
    ----------
    df : pd.DataFrame
        The data to divide
    path : str
        Path to the directory (possibly on GCS) in which to place the columns
    n_rgroup : int, default 1
        The number of row groups to partition the data into
        The rgroups will have approximately equal sizes
    rgroup_offset : int, default 0
        The index to start from in the name of file parts
        e.g. If rgroup_offset=10 then the first file will be `part_00010.pq`
    cgroup_columns : {cgroup: list of column names}
        The column lists to form cgroups; if None, do not make cgroups
        Each key is the name of the cgroup, and each value is the list of columns to include
        To reassemble later make sure to include join keys for each cgroup
    extension : str, default .pq
        The file extension for the dataframe (file type inferred from this extension
    convert : bool, default False
        If true attempt to coerce types to numeric. This can avoid issues with ambiguous
        object columns but requires additional time
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API
    write_args : dict
        Any additional args to pass to the write function

    """
    # Use a single dummy cgroup if None wanted
    if cgroup_columns is None:
        cgroup_columns = {None: df.columns}

    # Add leading dot if not in extension
    if extension[0] != ".":
        extension = "." + extension

    if convert:
        for col in df.columns:
            df[col] = pd.to_numeric(df[col], errors="ignore")

    for cname, columns in cgroup_columns.items():
        cgroup = df[columns]

        bucket = os.path.join(path, cname) if cname else path
        rnames = [
            "part_{:05d}{}".format(i + rgroup_offset, extension)
            for i in range(n_rgroup)
        ]
        if prefix is not None:
            rnames = [prefix + "_" + rn for rn in rnames]
        with filesystem.store(bucket, rnames) as datafiles:
            for rgroup, d in zip(np.array_split(cgroup, n_rgroup), datafiles):
                write_df(rgroup.reset_index(drop=True), d, **write_args)


def pickle(obj, path, filesystem=GCSNativeFileSystem()):
    """ Save a pickle of obj at the specified path

    Parameters
    ----------
    obj : Object
        Any pickle compatible object
    path : str
        The path to the location to save the pickle file, support gcs paths
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API
    """
    with filesystem.open(path, "wb") as f:
        _pickle.dump(obj, f)


def unpickle(path, filesystem=GCSNativeFileSystem()):
    """ Load an object from the pickle file at path

    Parameters
    ----------
    obj : Object
        Any pickle compatible object
    path : str
        The path to the location of the saved pickle file, support gcs paths
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API
    """
    with filesystem.open(path, "rb") as f:
        return _pickle.load(f)


def _collect(path, cgroups, rgroups, filesystem):
    """ Collect paths into cgroups and download any gcs files for local access

    Parameters
    ----------
    path : str
        The glob-able path to all datafiles to assemble into a frame
        e.g. gs://example/*/*, gs://example/*/part.0.pq, gs://example/c[1-2]/*
        See the README for a more detailed explanation
    cgroups : list of str, optional
        The list of cgroups (folder names) to include from the glob path
    rgroups : list of str, optional
        The list of rgroups (file names) to include from the glob path
    filesystem : blocks.filesystem.FileSystem or similar
        A filesystem object that implements the blocks.FileSystem API

    Returns
    -------
    grouped : {cgroup: list of paths}
        Paths to local copies of the data, grouped by cgroup

    """
    # ----------------------------------------
    # Collect paths into cgroups
    # ----------------------------------------
    paths = filesystem.ls(path)
    if not paths:
        raise ValueError("Did not find any files at the path: {}".format(path))
    expanded = _expand(paths, filesystem)
    filtered = _filter(expanded, cgroups, rgroups)
    grouped = _cgroups(filtered)
    accessed = _access(grouped, filesystem)

    # Go in specified cgroup order, or alphabetical if not specified
    if cgroups is None:
        cgroups = sorted(grouped.keys())

    return OrderedDict((k, accessed[k]) for k in cgroups)


def _has_ext(path):
    return os.path.splitext(path)[1] != ""


def _expand(paths, filesystem):
    """ For any directories in paths, expand into all the contained files
    """
    expanded = []
    for path in paths:
        if _has_ext(path):
            # Has an extension so treat it as a file
            expanded.append(path)
        else:
            # Otherwise try to read it like a directory
            expanded += filesystem.ls(os.path.join(path, "**"))
    return [p for p in expanded if _has_ext(p)]


def _filter(paths, cgroups, rgroups):
    """ Keep only paths with the appropriate cgroups and/or rgroups
    """
    kept = []
    for path in paths:
        valid_cgroup = cgroups is None or _cname(path) in cgroups
        valid_rgroup = rgroups is None or _rname(path) in rgroups
        if valid_cgroup and valid_rgroup:
            kept.append(path)
    return kept


def _base(path):
    """ Get base from path (name of the top level folder)
    """
    return os.path.dirname(os.path.dirname(path))


def _cname(path):
    """ Get cname from path (name of the parent folder)
    """
    return os.path.basename(os.path.dirname(path))


def _rname(path):
    """ Get cname from path (name of the file)
    """
    return os.path.basename(path)


def _cgroups(paths):
    """ Group paths by cgroup (the parent folder)
    """
    cgroups = defaultdict(list)
    for path in paths:
        cgroups[_cname(path)].append(path)
    return cgroups


def _access(cgroups, filesystem):
    """ Access potentially cloud stored files, preserving cgroups
    """
    updated = {}
    for cgroup, paths in cgroups.items():
        updated[cgroup] = filesystem.access(paths)
    return updated


def _safe_merge(df1, df2, merge="inner"):
    """ Merge two dataframes, warning of any shape differences
    """
    s1, s2 = df1.shape[0], df2.shape[0]
    if s1 != s2:
        warnings.warn(
            "The two cgroups have a different number of rows: {} {}".format(s1, s2)
        )
    return pd.merge(df1, df2, how=merge)


def _merge_all(frames, merge="inner"):
    """ Merge a list of dataframes with safe merge
    """
    result = frames[0]
    for frame in frames[1:]:
        result = _safe_merge(result, frame, merge)
    return result


def _shared_rgroups(grouped):
    rgroups = [[_rname(d.path) for d in group] for group in grouped.values()]
    return reduce(lambda a, b: set(a) & set(b), rgroups)



import atexit
import glob
import logging
import os
import shutil
import subprocess
import tempfile
from contextlib import contextmanager

from blocks.datafile import LocalDataFile
from blocks.filesystem.base import FileSystem
from six import string_types


class GCSFileSystem(FileSystem):
    """ File system interface that supports both local and GCS files

    This implementation uses subprocess and gsutil, which has excellent performance.
    However this can lead to problems in very multi-threaded applications and might not be
    as portable. For a python native implementation use GCSNativeFileSystem
    """

    GCS = "gs://"

    def __init__(self, parallel=True, quiet=True):
        flags = []
        if parallel:
            flags.append("-m")
        if quiet:
            flags.append("-q")
        self.gcscp = ["gsutil"] + flags + ["cp"]

    def local(self, path):
        """ Check if the path is available as a local file
        """
        return not path.startswith(self.GCS)

    def ls(self, path):
        """ List files correspond to path, including glob wildcards

        Parameters
        ----------
        path : str
            The path to the file or directory to list; supports wildcards
        """
        logging.info("Globbing file content in {}".format(path))
        if not self.local(path):
            with open(os.devnull, "w") as DEVNULL:
                p = subprocess.Popen(
                    ["gsutil", "ls", path],
                    stdout=subprocess.PIPE,
                    stderr=DEVNULL,
                    universal_newlines=True,
                )
                stdout = p.communicate()[0]
            output = [line for line in stdout.split("\n") if line and line[-1] != ":"]
        elif "**" in path:
            # Manual recursive glob, since in 2.X glob doesn't have recursive support
            path = path.rstrip("*")
            output = []
            for root, subdirs, files in os.walk(path):
                for fname in files:
                    output.append(os.path.join(root, fname))
        elif os.path.isdir(path):
            output = [os.path.join(path, f) for f in os.listdir(path)]
        else:
            output = glob.glob(path)
        return sorted(p.rstrip("/") for p in output)

    def rm(self, paths, recursive=False):
        """ Remove the files at paths

        Parameters
        ----------
        paths : list of str
            The paths to remove
        recursive : bool, default False
            If true, recursively remove any directories
        """
        if isinstance(paths, string_types):
            paths = [paths]

        if any(not self.local(p) for p in paths):
            # at least one location is on GCS
            cmd = ["gsutil", "-m", "rm"]

        else:
            cmd = ["rm"]

        if recursive:
            cmd.append("-r")

        CHUNK_SIZE = 1000
        paths_chunks = [
            paths[x : x + CHUNK_SIZE] for x in range(0, len(paths), CHUNK_SIZE)
        ]
        for paths in paths_chunks:
            subprocess.check_call(cmd + paths)

    def cp(self, sources, dest, recursive=False):
        """ Copy the files in sources to dest

        Parameters
        ----------
        sources : list of str
            The list of paths to copy
        dest : str
            The destination for the copy of source(s)
        recursive : bool
            If true, recursively copy any directories
        """
        if isinstance(sources, string_types):
            sources = [sources]

        summary = ", ".join(sources)
        logging.info("Copying {} to {}...".format(summary, dest))

        if any(self.GCS in x for x in sources + [dest]):
            # at least one location is on GCS
            cmd = self.gcscp
        else:
            cmd = ["cp"]

        if recursive:
            cmd.append("-r")

        CHUNK_SIZE = 1000
        sources_chunks = [
            sources[x : x + CHUNK_SIZE] for x in range(0, len(sources), CHUNK_SIZE)
        ]
        for sources in sources_chunks:
            subprocess.check_call(cmd + sources + [dest])

    @contextmanager
    def open(self, path, mode="rb"):
        """ Access path as a file-like object

        Parameters
        ----------
        path: str
            The path of the file to access
        mode: str
            The file mode for the opened file

        Returns
        -------
        file: file
            A python file opened to the provided path (uses a local temporary copy that is removed)
        """
        with tempfile.NamedTemporaryFile() as nf:
            if mode.startswith("r"):
                self.cp(path, nf.name)

            nf.seek(0)

            with open(nf.name, mode) as f:
                yield f

            nf.seek(0)

            if mode.startswith("w"):
                self.cp(nf.name, path)

    def access(self, paths):
        """ Access multiple paths as file-like objects

        This allows for optimization like parallel downloads

        Parameters
        ----------
        paths: list of str
            The paths of the files to access

        Returns
        -------
        files: list of DataFile
            A list of datafile instances, one for each input path
        """
        # Move the files into a tempdir from GCS
        tmpdir = _session_tempdir()
        self.cp(paths, tmpdir, recursive=True)

        # Then get file handles for each
        datafiles = []
        for path in paths:
            local = os.path.join(tmpdir, os.path.basename(path))
            datafiles.append(LocalDataFile(path, local))
        return datafiles

    @contextmanager
    def store(self, bucket, files):
        """ Create file stores that will be written to the filesystem on close

        This allows for optimizations when storing several files

        Parameters
        ----------
        bucket : str
            The path of the bucket (on GCS) or folder (local) to store the data in
        files : list of str
            The filenames to create

        Returns
        -------
        datafiles : contextmanager
            A context manager that yields datafiles and when the context is closed
            they are written to GCS

        Usage
        -----
        >>> with filesystem.store('gs://bucket/sub/', ['ex1.txt', 'ex2.txt']) as datafiles:
        >>>     datafiles[0].handle.write('example 1')
        >>>     datafiles[1].handle.write('example 2')
        """
        # Make local files in a tempdir that serve as the file handles
        tmpdir = _session_tempdir()
        datafiles = []
        local_files = []
        for f in files:
            local = os.path.join(tmpdir, f)
            local_files.append(local)
            datafiles.append(LocalDataFile(os.path.join(bucket, f), local))

        yield datafiles

        if self.local(bucket) and not os.path.exists(bucket):
            os.makedirs(bucket)

        self.cp(local_files, os.path.join(bucket, ""), recursive=True)


def _session_tempdir():
    """ Create a tempdir that will be cleaned up at session exit
    """
    tmpdir = tempfile.mkdtemp()
    # create and use a subdir of specified name to preserve cgroup logic
    atexit.register(lambda: shutil.rmtree(tmpdir))
    return tmpdir



import json
import os
import pandas as pd
import numpy as np
import pytest
import six
import blocks.dfio as io
from blocks.datafile import LocalDataFile


def test_read_write_native_formats(randomdata, datadir_local):
    extensions = [".pkl", ".csv", ".json", ".json.gz", ".pkl.gz", ".csv.gz"]
    for extension in extensions:
        path = os.path.join(datadir_local, "tmp{}".format(extension))
        df = _reload(randomdata, path)
        assert np.isclose(df, randomdata).all().all()


def test_compression_csv(randomdata, datadir_local):
    path = os.path.join(datadir_local, "tmp.csv.gz")
    # write compressed with pandas
    randomdata.to_csv(path, index=False, compression="gzip")

    # read compressed with blocks
    d = LocalDataFile(path, path)
    df = io.read_df(d)
    assert np.isclose(df, randomdata).all().all()

    # write compressed with blocks
    d = LocalDataFile(path, path)
    io.write_df(randomdata, d)
    # read compressed with pandas
    df = pd.read_csv(path, compression="gzip")
    assert np.isclose(df, randomdata).all().all()


def test_compression_pickle(randomdata, datadir_local):
    path = os.path.join(datadir_local, "tmp.pkl.gz")
    # write compressed with pandas
    randomdata.to_pickle(path, compression="gzip")

    # read compressed with blocks
    d = LocalDataFile(path, path)
    df = io.read_df(d)
    assert np.isclose(df, randomdata).all().all()

    # write compressed with blocks
    d = LocalDataFile(path, path)
    io.write_df(randomdata, d)
    # read compressed with pandas
    df = pd.read_pickle(path, compression="gzip")
    assert np.isclose(df, randomdata).all().all()


def test_compression_parquet(randomdata, datadir_local):
    pytest.importorskip("pyarrow")
    pytest.importorskip("pandas", minversion="0.22.0")
    path = os.path.join(datadir_local, "tmp.parquet.gz")
    # write compressed with pandas
    randomdata.to_parquet(path, compression="gzip")

    # read compressed with blocks
    d = LocalDataFile(path, path)
    df = io.read_df(d)
    assert np.isclose(df, randomdata).all().all()

    # write compressed with blocks
    d = LocalDataFile(path, path)
    io.write_df(randomdata, d)
    # read compressed with pandas
    df = pd.read_parquet(path)
    assert np.isclose(df, randomdata).all().all()


def test_read_write_parquet(randomdata, datadir_local):
    pytest.importorskip("pyarrow")
    pytest.importorskip("pandas", minversion="0.22.0")
    path = os.path.join(datadir_local, "tmp.pq")
    df = _reload(randomdata, path)
    assert np.isclose(df, randomdata).all().all()


def test_read_write_parquet_unicode(randomdata, datadir_local):
    pytest.importorskip("pyarrow")
    pytest.importorskip("pandas", minversion="0.22.0")
    randomdata[u"f10"] = randomdata["f9"]
    path = os.path.join(datadir_local, "tmp{}".format(".pq"))
    if six.PY2:
        # expect a warning about this in python2
        with pytest.warns(UserWarning):
            df = _reload(randomdata, path)
    else:
        df = _reload(randomdata, path)
    assert np.isclose(df, randomdata).all().all()


def test_read_write_avro(randomdata, datadir_local):
    pytest.importorskip("fastavro")
    path = os.path.join(datadir_local, "tmp.avro")
    schema_path = os.path.join(datadir_local, "tmp.avsc")
    _write_schema(randomdata, schema_path)
    df = _reload(randomdata, path, schema_path=schema_path)
    assert np.isclose(df, randomdata).all().all()


def _write_schema(data, path):
    schema = {"type": "record", "name": "testschema"}
    schema["fields"] = [{"name": str(c), "type": "float"} for c in data.columns.values]
    with open(path, "w") as f:
        json.dump(schema, f)


def _reload(data, path, **kwargs):
    d = LocalDataFile(path, path)
    io.write_df(data, d, **kwargs)
    d = LocalDataFile(path, path)
    df = io.read_df(d)
    return df



import os
import numpy as np
import blocks


def test_place(datadir, randomdata, fs):
    blocks.place(randomdata, os.path.join(datadir, "example.csv"))
    assert len(fs.ls(datadir)) == 1
    df = blocks.assemble(datadir)
    assert np.isclose(df, randomdata).all().all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from blocks import core


def test_assemble(populated):
    df = core.assemble(populated)
    assert df.shape == (40, 41)
    expected = ["f{}_{}".format(i, j) for i in range(4) for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)


def test_assemble_ordered(populated):
    order = ["c2", "c1", "c3", "c0"]
    df = core.assemble(populated, cgroups=order)
    assert df.shape == (40, 41)
    expected = ["f{}_{}".format(order[i][1], j) for i in range(4) for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)
    # Check the features are in the right order
    assert [c for c in df.columns if c != "key"] == expected[:-1]


# Various options do not depend on filesystem so we can just test locally
def test_assemble_filtered_cgroup(populated_local, keys):
    df = core.assemble(populated_local, cgroups=["c0", "c3"])
    assert df.shape == (40, 21)
    expected = ["f{}_{}".format(i, j) for i in [0, 3] for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)
    assert (df.key == keys).all()


def test_assemble_filtered_rgroup(populated_local, keys):
    df = core.assemble(populated_local, rgroups=["part.0.csv", "part.1.csv"])
    assert df.shape == (20, 41)
    expected = ["f{}_{}".format(i, j) for i in range(4) for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)
    assert (df.key == keys[:20].reset_index(drop=True)).all()


def test_assemble_read_args(populated_local, keys):
    read_args = {"dtype": str}
    df = core.assemble(populated_local, read_args=read_args)
    assert df.shape == (40, 41)
    expected = ["f{}_{}".format(i, j) for i in range(4) for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)
    assert (df.key == keys).all()
    assert (df.dtypes == "object").all()


def test_assemble_cgroup_args(populated_local, keys):
    cgroup_args = {"c0": {"dtype": str}}
    df = core.assemble(populated_local, cgroup_args=cgroup_args)
    assert df.shape == (40, 41)
    expected = ["f{}_{}".format(i, j) for i in range(4) for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)
    assert (df.key == keys).all()
    for col in ["f0_{}".format(i) for i in range(10)]:
        assert df.dtypes[col] == "object"



import os
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager

from six import add_metaclass


@add_metaclass(ABCMeta)
class DataFile:
    """ A datafile that holds the source path and method to get a handle
    """

    def __init__(self, path):
        self.path = path

    @abstractmethod
    def handle(self):
        pass


class LocalDataFile(DataFile):
    def __init__(self, path, local):
        self.local = local
        super(LocalDataFile, self).__init__(path)

    @contextmanager
    def handle(self, mode="rb"):
        if mode.startswith("w") and not os.path.isdir(os.path.dirname(self.local)):
            os.makedirs(os.path.dirname(self.local))
        with open(self.local, mode) as f:
            yield f



from abc import ABCMeta, abstractmethod

from six import add_metaclass


@add_metaclass(ABCMeta)
class FileSystem(object):
    """ The required interface for any filesystem implementation

    See GCSFileSystem for a full implementation. This FileSystem is intended
    to be extendable to support cloud file systems, encryption strategies, etc...
    """

    @abstractmethod
    def ls(self, path):
        """ List files correspond to path, including glob wildcards

        Parameters
        ----------
        path : str
            The path to the file or directory to list; supports wildcards
        """
        pass

    @abstractmethod
    def access(self, paths):
        """ Access multiple paths as file-like objects

        This allows for optimization like parallel downloads

        Parameters
        ----------
        paths: list of str
            The paths of the files to access

        Returns
        -------
        files: list of DataFile
            A list of datafile instances, one for each input path
        """
        pass

    @abstractmethod
    def store(self, bucket, files):
        """ Store multiple data objects

        This allows for optimizations when storing several files

        Parameters
        ----------
        bucket : str
            The GCS bucket to use to store the files
        files : list of str
            The file names to store

        Returns
        -------
        datafiles : contextmanager
           A contextmanager that will yield datafiles and place them
           on the filesystem when finished
        """
        pass



# -*- coding: utf-8 -*-

__author__ = """Bradley Axen"""
__email__ = "baxen@squareup.com"
__version__ = "0.7.1"


from blocks.core import (
    assemble,
    iterate,
    partitioned,
    divide,
    place,
    pickle,
    unpickle,
)  # NOQA
from blocks.dfio import read_df, write_df  # NOQA



import fnmatch
import logging
import os
import shutil
import time
from contextlib import contextmanager
from io import BytesIO, TextIOWrapper

import requests
import wrapt
from blocks.datafile import DataFile, LocalDataFile
from blocks.filesystem.gcs_filesystem import GCSFileSystem
from google.cloud import storage
from six import string_types


class GCSNativeDataFile(DataFile):
    def __init__(self, path, filesystem):
        self.filesystem = filesystem
        super(GCSNativeDataFile, self).__init__(path)

    @contextmanager
    def handle(self, mode="rb"):
        bytesio = BytesIO()

        if mode.startswith("r"):
            self.filesystem._read(self.path, bytesio)

        if mode.endswith("b"):
            buf = bytesio
        else:
            buf = TextIOWrapper(bytesio)

        yield buf

        buf.seek(0)

        if mode.startswith("w"):
            self.filesystem._write(self.path, bytesio)


@wrapt.decorator
def _retry_with_backoff(wrapped, instance, args, kwargs):
    trial = 0
    while True:
        wait = 2 ** (trial + 2)  # 4s up to 128s
        try:
            return wrapped(*args, **kwargs)
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ChunkedEncodingError,
        ):
            if trial == 6:
                raise
            logging.info(
                "{} failed to connect, retrying after {}s".format(
                    wrapped.__name__, wait
                )
            )
            trial += 1
        time.sleep(wait)


class GCSNativeFileSystem(GCSFileSystem):
    """ File system interface that supports GCS and local files

    This uses the native python cloud storage library for read and write, rather than gsutil.
    The performance is significantly slower when doing any operations over several files (especially
    copy), but is thread-safe for applications which are already parallelized. It stores the files
    entirely in memory rather than using tempfiles.
    """

    def __init__(self, *args, **kwargs):
        self._client = None
        super(GCSNativeFileSystem, self).__init__(*args, **kwargs)

    def client(self):
        # Load client only when needed, so that this can be used for local paths without connecting
        if self._client is None:
            self._client = storage.Client()
        return self._client

    def ls(self, path):
        """ List all files at the specified path, supports globbing
        """
        logging.info("Globbing file content in {}".format(path))

        # use GCSFileSystem's implementation for local paths
        if self.local(path):
            return super(GCSNativeFileSystem, self).ls(path)

        # find all names that start with the prefix for this path
        bucket, path = self._split(path)
        prefix = self._prefix(path)

        if prefix == path:
            # no pattern matching
            iterator = self._list_blobs(bucket, prefix=prefix, delimiter="/")
            names = [b.name for b in iterator]
            names += iterator.prefixes
            # if we ran ls('gs://bucket/dir') we need to rerun with '/' to get the content
            if names == [os.path.join(path, "")]:
                return self.ls(os.path.join("gs://" + bucket, path, ""))
        else:
            # we have a pattern to match
            names = [b.name for b in self._list_blobs(bucket, prefix=prefix)]
            # for recursive glob, do not attempt to match folders, only files
            if "**" in path:
                names = fnmatch.filter(names, path)
            else:
                # make sure we can match folders in addition to blobs
                candidates = set(names) | set(self._list_single(prefix, names))
                names = fnmatch.filter(candidates, path)
                # one more filter because fnmatch recurses single *
                names = [name for name in names if name.count("/") == path.count("/")]

        paths = ["gs://{}/{}".format(bucket, name) for name in names]
        return sorted(p.rstrip("/") for p in paths)

    def is_dir(self, path):
        # Check if a path is a directory, locally or on GCS
        if self.local(path):
            return os.path.isdir(path)
        elif path.endswith("/"):
            return True
        else:
            return self.ls(path) != [path]

    def copy_single(self, source, dest):
        local_source = self.local(source)
        local_dest = self.local(dest)

        if local_source and local_dest:
            shutil.copy(source, dest)

        if not local_source and local_dest:
            if os.path.isdir(dest):
                dest = os.path.join(dest, os.path.basename(source))
            self._blob(source).download_to_filename(dest)

        if local_source and not local_dest:
            if dest.endswith("/"):
                dest = os.path.join(dest, os.path.basename(source))
            self._blob(dest).upload_from_filename(source)

        if not local_source and not local_dest:
            if dest.endswith("/"):
                dest = os.path.join(dest, os.path.basename(source))
            self._transfer(source, dest)

    def cp(self, sources, dest, recursive=False):
        """ Copy the files in sources (recursively) to dest

        Parameters
        ----------
        sources : list of str
            The list of paths to copy, which can be directories
        dest : str
            The destination for the copy of source(s)
        recursive : bool, default False
            If true, recursively copy directories
        """
        if isinstance(sources, string_types):
            sources = [sources]

        for source in sources:
            if recursive and self.is_dir(source):
                # Note: if source ends with a '/', this copies the content into dest
                #   and if source does not, this copies the whole directory into dest
                #   this is the same behavior as copy
                subsource = [s.rstrip("/") for s in self.ls(source)]
                subdest = os.path.join(dest, os.path.basename(source), "")
                if self.local(subdest) and not os.path.exists(subdest):
                    os.makedirs(subdest)
                self.cp(subsource, subdest, recursive=True)
            else:
                self.copy_single(source, dest)

    def rm_single(self, path):
        if self.local(path):
            os.remove(path)
        else:
            self._blob(path).delete()

    def rm(self, paths, recursive=False):
        """ Remove the files at paths

        Parameters
        ----------
        paths : list of str
            The paths to remove
        recursive : bool, default False
            If true, recursively remove any directories
        """
        if isinstance(paths, string_types):
            paths = [paths]

        for path in paths:
            if recursive and self.is_dir(path) and not self.local(path):
                self.rm(self.ls(path), recursive=True)
            elif recursive and self.is_dir(path):
                shutil.rmtree(path)
            else:
                self.rm_single(path)

    def datafile(self, path):
        if self.local(path):
            return LocalDataFile(path, path)
        return GCSNativeDataFile(path, self)

    @contextmanager
    def open(self, path, mode="rb"):
        """ Access paths as a file-like object

        Parameters
        ----------
        path: str
            The path of the file to access
        mode: str
            The file mode for the opened file

        Returns
        -------
        file: BytesIO
            A BytesIO handle for the specified path, works like a file object
        """
        with self.datafile(path).handle(mode) as handle:
            yield handle

    def access(self, paths):
        """ Access multiple paths as file-like objects

        This allows for optimization like parallel downloads. To help track which files
        came from which objects, this returns instances of Datafile

        Parameters
        ----------
        paths: list of str
            The paths of the files to access

        Returns
        -------
        files: list of DataFile
            A list of datafile instances, one for each input path
        """
        datafiles = []
        for path in paths:
            datafiles.append(self.datafile(path))
        return datafiles

    @contextmanager
    def store(self, bucket, files):
        """ Create file stores that will be written to the filesystem on close

        This allows for optimizations when storing several files

        Parameters
        ----------
        bucket : str
            The path of the bucket (on GCS) or folder (local) to store the data in
        files : list of str
            The filenames to create

        Returns
        -------
        datafiles : contextmanager
            A context manager that yields datafiles and when the context is closed
            they are written to GCS

        Usage
        -----
        >>> with filesystem.store('gs://bucket/sub/', ['ex1.txt', 'ex2.txt']) as datafiles:
        >>>     datafiles[0].handle.write('example 1')
        >>>     datafiles[1].handle.write('example 2')
        """
        # Make BytesIO instances that serve as the file handles
        datafiles = []
        for f in files:
            datafiles.append(self.datafile(os.path.join(bucket, f)))

        yield datafiles

    def _prefix(self, path):
        # find the lowest prefix that doesn't include a pattern match
        splits = path.split("/")
        accumulate = []
        while splits:
            sub = splits.pop(0)
            if any(x in sub for x in ["*", "?", "[", "]"]):
                break
            accumulate.append(sub)
        return "/".join(accumulate) if accumulate else None

    def _list_single(self, prefix, names):
        # find files that are directly in the dir specified by prefix, not in subfolders
        valid = set(n.replace(prefix, "").lstrip("/").split("/")[0] for n in names)
        return [os.path.join(prefix, n) for n in valid]

    def _split(self, path):
        bucket = path.replace(self.GCS, "").split("/")[0]
        prefix = "gs://{}".format(bucket)
        path = path[len(prefix) + 1 :]
        return bucket, path

    @_retry_with_backoff
    def _list_blobs(self, bucket, prefix=None, delimiter=None):
        return (
            self.client()
            .get_bucket(bucket)
            .list_blobs(prefix=prefix, delimiter=delimiter)
        )

    @_retry_with_backoff
    def _blob(self, path):
        bucket, path = self._split(path)
        return storage.Blob(path, self.client().get_bucket(bucket))

    @_retry_with_backoff
    def _transfer(self, path1, path2):
        bucket1, path1 = self._split(path1)
        bucket2, path2 = self._split(path2)
        source_bucket = self.client().get_bucket(bucket1)
        source_blob = source_bucket.blob(path1)
        destination_bucket = self.client().get_bucket(bucket2)
        source_bucket.copy_blob(source_blob, destination_bucket, path2)

    @_retry_with_backoff
    def _read(self, path, bytesio):
        self._blob(path).download_to_file(bytesio)
        bytesio.seek(0)

    @_retry_with_backoff
    def _write(self, path, bytesio):
        bytesio.seek(0)
        self._blob(path).upload_from_file(bytesio)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from blocks import core


def test_partitioned(populated):
    pytest.importorskip("dask.dataframe")
    df = core.partitioned(populated).compute()
    assert df.shape == (40, 41)
    expected = ["f{}_{}".format(i, j) for i in range(4) for j in range(10)]
    expected.append("key")
    assert set(df.columns) == set(expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from blocks import core


def test_expand(populated, fs):
    expected = [
        os.path.join(populated, "c{}/part.{}.csv".format(i, j))
        for i in range(4)
        for j in range(4)
    ]

    # All of these patterns should expand into the same set of files
    for ex in ["", "*", "*/*", "**"]:
        ex = os.path.join(populated, ex)
        paths = fs.ls(ex)
        expanded = sorted(core._expand(paths, fs))
        assert expanded == expected


def test_expand_pattern(populated, fs):
    expected = [
        os.path.join(populated, "c{}/part.{}.csv".format(i, j))
        for i in range(2)
        for j in range(4)
    ]

    # All of these patterns should expand into the same set of files
    for ex in ["c[01]", "c[01]/*"]:
        ex = os.path.join(populated, ex)
        paths = fs.ls(ex)
        expanded = sorted(core._expand(paths, fs))
        assert expanded == expected


def test_cgroups():
    expanded = ["base/c{}/part.{}.csv".format(i, j) for i in range(4) for j in range(4)]
    cgroups = core._cgroups(expanded)
    for i in range(4):
        key = "c{}".format(i)
        assert key in cgroups
        assert cgroups[key] == ["base/c{}/part.{}.csv".format(i, j) for j in range(4)]


def test_access(populated, fs):
    paths = fs.ls(populated)
    expanded = core._expand(paths, fs)
    cgroups = core._cgroups(expanded)
    cgroups = core._access(cgroups, fs)
    assert len(cgroups) == 4
    for c, datafiles in cgroups.items():
        assert len(datafiles) == 4
        for d in datafiles:
            with d.handle() as f:
                assert f.read()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "six",
    "wrapt",
    "pandas",
    "google-cloud-storage",
]

extras_require = {
    "pq": ["pyarrow"],
    "avro": ["fastavro"],
    "tests": ["pytest", "pytest-cov", "delegator.py", "flake8"],
    "doc": ["sphinx", "numpydoc", "sphinx-rtd-theme"],
    "format": ["pre-commit"],
}
extras_require["dev"] = set(sum(extras_require.values(), []))

setup(
    name="sq-blocks",
    version="0.7.1",
    description=(
        "Blocks provides a simple interface to read, organize, and manipulate structured data"
        " in files on local and cloud storage"
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Bradley Axen",
    author_email="baxen@squareup.com",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    zip_safe=False,
    keywords="blocks",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ],
)
